# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for capacity.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,unused-argument
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import capacity
from ..air_sdk.simulation import Simulation

class TestCapacity(TestCase):
    def setUp(self):
        self.model = capacity.Capacity(MagicMock())
        self.model.copies = 30

    def test_init_(self):
        self.assertFalse(self.model._deletable)
        self.assertFalse(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Capacity {self.model.copies}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

class TestCapacityApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = capacity.CapacityApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/capacity/')

    @patch('air_sdk.air_sdk.capacity.CapacityApi.get')
    def test_get_capacity_by_sim(self, mock_get):
        mock_sim = MagicMock()
        res = self.api.get_capacity(mock_sim)
        mock_get.assert_called_with(simulation_id=mock_sim.id)
        self.assertEqual(res, mock_get.return_value)

    @patch('air_sdk.air_sdk.capacity.CapacityApi.get')
    def test_get_capacity_by_id(self, mock_get):
        res = self.api.get_capacity(simulation_id='abc123')
        mock_get.assert_called_with(simulation_id='abc123')
        self.assertEqual(res, mock_get.return_value)

    def test_get_capacity_missing_param(self):
        with self.assertRaises(ValueError) as err:
            self.api.get_capacity()
        self.assertEqual(str(err.exception), 'Must pass a simulation or simulation_id argument')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/capacity/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, capacity.Capacity)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get_simulation(self, mock_raise):
        sim = Simulation(MagicMock())
        sim.id = 'abc123'
        self.api.get(sim)
        self.client.get.assert_called_with(f'{self.client.api_url}/capacity/abc123/', params={})



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Node module
"""

from . import util
from .air_model import AirModel

class Node(AirModel):
    """
    Manage a Node

    ### delete
    Delete the node. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the node

    ### refresh
    Syncs the node with all values returned by the API

    ### update
    Update the node with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _ignored_update_fields = ['interfaces']

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Node {self.name} {self.id}>'

class NodeApi:
    """ High-level interface for the Node API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/node/'

    @util.deprecated('NodeApi.list()')
    def get_nodes(self, simulation_id=''): #pylint: disable=missing-function-docstring
        return self.list(simulation=simulation_id)

    def get(self, node_id, **kwargs):
        """
        Get an existing node

        Arguments:
            node_id (str): Node ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Node`](/docs/node)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.nodes.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Node server 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        if kwargs.get('simulation_id'):
            kwargs['simulation'] = kwargs['simulation_id']
            del kwargs['simulation_id']
        url = f'{self.url}{node_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Node(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing nodes

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.nodes.list()
        [<Node server c51b49b6-94a7-4c93-950c-e7fa4883591>, <Node switch 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Node(self, **node) for node in res.json()]

    @util.required_kwargs(['name', 'topology'])
    def create(self, **kwargs):
        """
        Create a new node

        Arguments:
            name (str): Node name
            topology (str | `Topology`): `Topology` or ID
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Node`](/docs/node)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.nodes.create(name='server', topology=topology)
        <Node server 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Node(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Token module
"""

from . import util
from .air_model import AirModel

class Token(AirModel):
    """
    View an API Token

    ### json
    Returns a JSON string representation of the interface

    ### refresh
    Syncs the interface with all values returned by the API
    """
    #_deletable = False
    _updatable = False

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        if hasattr(self, 'id'):
            return f'<Token {self.name} {self.id}>'
        if hasattr(self, 'token'):
            return f'<Token {self.name} {self.token}>'
        return f'<Token {self.name}>'

class TokenApi:
    """ High-level interface for the Token API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/api-token/'

    def delete(self, token_id, **kwargs):
        """
        Deletes an api token

        Arguments:
            token_id (str): Token ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters


        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 204 No Content
            or valid response JSON

        Example:
        ```
        >>> air.api_tokens.delete('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        ```
        """
        url = f'{self.url}{token_id}/'
        res = self.client.delete(url, params=kwargs)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing tokens

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.tokens.list()
        [<Interface eth0 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Interface eth1 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """
        #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Token(self, **token) for token in res.json()]

    @util.required_kwargs(['name'])
    def create(self, **kwargs):
        """
        Add a new api token to your account

        Arguments:
            name (str): Descriptive name for the api token
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`api-token`](/docs/api-token)

        Raises:
        [`AirUnexpectedResponse`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.api_tokens.create(name='my_api_token')
        <Token my_api_token 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Token(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Capacity module
"""

from . import util
from .air_model import AirModel

class Capacity(AirModel):
    """
    View platform capacity

    ### json
    Returns a JSON string representation of the capacity

    ### refresh
    Syncs the capacity with all values returned by the API
    """
    _deletable = False
    _updatable = False

    def __repr__(self):
        if self._deleted or not self.copies:
            return super().__repr__()
        return f'<Capacity {self.copies}>'

class CapacityApi:
    """ High-level interface for the Simulation API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/capacity/'

    @util.deprecated('CapacityApi.get()')
    def get_capacity(self, simulation=None, simulation_id=None): #pylint: disable=missing-function-docstring
        if not simulation and not simulation_id:
            raise ValueError('Must pass a simulation or simulation_id argument')
        sim_id = simulation_id or simulation.id
        return self.get(simulation_id=sim_id)

    def get(self, simulation_id, **kwargs):
        """
        Get current platform capacity for a [`Simulation`](/docs/simulation)

        Arguments:
            simulation_id (str | `Simulation`): Simulation or ID

        Returns:
        [`Capacity`](/docs/capacity)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.capacity.get(simulation)
        <Capacity 30>
        ```
        """
        if isinstance(simulation_id, AirModel):
            simulation_id = simulation_id.id
        url = f'{self.url}{simulation_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Capacity(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for organization.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,protected-access
from copy import deepcopy
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import organization

class TestOrganization(TestCase):
    def setUp(self):
        self.api = MagicMock()
        mock_res = MagicMock()
        self.api.client.post.return_value = mock_res
        self.model = organization.Organization(self.api, id='abc123', name='NVIDIA')
        self.model._api_url = '/organization/'

    def test_init(self):
        self.assertTrue(self.model._deletable)
        self.assertTrue(self.model._updatable)
        self.assertEqual(self.model.ORG_MEMBER_ROLE, 'Organization Member')
        self.assertEqual(self.model._members_api_url,
                         f'{self.api.url}{self.model.id}/members/')

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Organization {self.model.name} {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    def test_add_member(self, mock_refresh):
        username = 'user1@nvidia.com'
        self.model.add_member(username)
        self.api.client.post.assert_called_once_with(self.model._members_api_url,
                                                     json={'username': username,
                                                           'roles': [self.model.ORG_MEMBER_ROLE]})
        mock_refresh.assert_called_once()

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    def test_add_member_with_role(self, mock_refresh):
        username = 'user1@nvidia.com'
        role = 'test role'
        self.model.add_member(username, [role])
        self.api.client.post.assert_called_once_with(self.model._members_api_url,
                                                     json={'username': username, 'roles': [role]})
        mock_refresh.assert_called_once()

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    def test_add_members(self, mock_refresh):
        member1 = {'username': 'user1@nvidia.com', 'roles': ['Organization Admin']}
        member2 = {'username': 'user2@nvidia.com'}
        member2_with_role = deepcopy(member2)
        member2_with_role['roles'] = [self.model.ORG_MEMBER_ROLE]
        self.model.add_members([member1, member2])
        self.api.client.post.assert_called_once_with(self.model._members_api_url,
                                                     json=[member1, member2_with_role])
        mock_refresh.assert_called_once()

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    def test_remove_member(self, mock_refresh):
        username = 'user1@nvidia.com'
        self.model.remove_member(username)
        self.api.client.delete.assert_called_once_with(self.model._members_api_url,
                                                       json={'username': username})
        mock_refresh.assert_called_once()
        mock_refresh.assert_called_once()

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    def test_remove_member_no_refresh(self, mock_refresh):
        username = 'user1@nvidia.com'
        self.model.remove_member(username, _refresh_when_done=False)
        self.api.client.delete.assert_called_once()
        mock_refresh.assert_not_called()

    @patch('air_sdk.air_sdk.organization.Organization.refresh')
    @patch('air_sdk.air_sdk.organization.Organization.remove_member')
    def test_remove_members(self, mock_remove, mock_refresh):
        members = ['user1@nvidia.com', 'user2@nvidia.com']
        self.model.remove_members(members)
        mock_for_assert = MagicMock()
        mock_for_assert(members[0], _refresh_when_done=False)
        mock_for_assert(members[1], _refresh_when_done=False)
        self.assertEqual(mock_remove.mock_calls, mock_for_assert.mock_calls)
        mock_refresh.assert_called_once()

class TestOrganizationApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = organization.OrganizationApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/organization/')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        org_id = 'abc123'
        self.client.get.return_value.json.return_value = {'id': org_id, 'name': 'test'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/organization/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, organization.Organization)
        self.assertEqual(res.id, org_id)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/organization/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], organization.Organization)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create(self, mock_raise):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(name='abc123', members=['def123'])
        self.client.post.assert_called_with(f'{self.client.api_url}/organization/',
                                            json={'name': 'abc123', 'members': ['def123']})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, organization.Organization)
        self.assertEqual(res.id, 'abc')

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create(members=[])
        self.assertTrue('requires name' in str(err.exception))



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Marketplace Demo module
"""

from . import util
from .air_model import AirModel

class Marketplace(AirModel):
    """
    Manage marketplace demos

    """
    _updatable = False
    _deletable = False

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Marketplace Demo \'{self.name}\' {self.id}>'

class MarketplaceApi:
    """ High-level interface for the Marketplace API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/marketplace/demo/'

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing keys

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.marketplace.list()
        [<Marketplace Demo EVPN Centralized c51b49b6-94a7-4c93-950c-e7fa4883591>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Marketplace(self, **key) for key in res.json()]

    def get(self, demo_id, **kwargs):
        """
        Get an existing marketplace demo

        Arguments:
            demo_id (str): Demo ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Demo`](/docs/marketplace)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.marketplace.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Marketplace Demo EVPN Centralized 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{demo_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Marketplace(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Custom exceptions for the AIR SDK
"""
class AirError(Exception):
    """
    Base exception class. All custom exceptions should inherit from this class.
    """
    def __init__(self, message='', status_code=None):
        self.status_code = status_code
        super().__init__(message)

class AirAuthorizationError(AirError):
    """ Raised when authorization with the API fails. """
    def __init__(self, message='An error occurred when authorizing the Air API', status_code=None):
        self.message = message
        super().__init__(message=self.message, status_code=status_code)

class AirUnexpectedResponse(AirError):
    """ Raised when the API returns an unexpected response. """
    def __init__(self, message='', status_code=None):
        self.message = 'Received an unexpected response from the Air API'
        if status_code:
            self.message += f' ({status_code})'
        self.message += f': {message}'
        super().__init__(message=self.message, status_code=status_code)

class AirForbiddenError(AirError):
    """ Raised when an API call returns a 403 Forbidden error """
    def __init__(self, message='Received 403 Forbidden. Please call AirApi.authorize().'):
        self.message = message
        super().__init__(message=self.message, status_code=403)

class AirObjectDeleted(AirError):
    """ Raised when accessing a previously instantiated object that has since been deleted """
    def __init__(self, cls, message=''):
        self.message = message
        if not self.message:
            self.message = f'{cls} object has been deleted and should no longer be referenced'
        super().__init__(message=self.message)



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Helper utils
"""

import datetime
import logging
from json import JSONDecodeError

from dateutil import parser as dateparser

from .exceptions import AirUnexpectedResponse

def raise_if_invalid_response(res, status_code=200, data_type=dict):
    """
    Validates that a given API response has the expected status code and JSON payload

    Arguments:
    res (requests.HTTPResponse) - API response object
    status_code [int] - Expected status code (default: 200)

    Raises:
    AirUnexpectedResponse - Raised if an unexpected response is received from the API
    """
    json = None
    if res.status_code != status_code:
        logging.debug(res.text)
        raise AirUnexpectedResponse(message=res.text, status_code=res.status_code)
    if not data_type:
        return
    try:
        json = res.json()
    except JSONDecodeError:
        raise AirUnexpectedResponse(message=res.text, status_code=res.status_code)
    if not isinstance(json, data_type):
        raise AirUnexpectedResponse(message=f'Expected API response to be of type {data_type}, ' + \
                                            f'got {type(json)}',
                                    status_code=res.status_code)

def required_kwargs(required):
    """ Decorator to enforce required kwargs for a function """
    if not isinstance(required, list):
        required = [required]
    def wrapper(method):
        def wrapped(*args, **kwargs):
            for arg in required:
                if isinstance(arg, tuple):
                    present = False
                    for option in arg:
                        if option in kwargs:
                            present = True
                            break
                    if not present:
                        raise AttributeError(f'{method} requires one of the following: {arg}')
                else:
                    if arg not in kwargs:
                        raise AttributeError(f'{method} requires {arg}')
            return method(*args, **kwargs)
        return wrapped
    return wrapper

def deprecated(new=None):
    """ Decorator to log a warning when calling a deprecated function """
    def wrapper(method):
        def wrapped(*args, **kwargs):
            msg = f'{method} has been deprecated and will be removed in a future release.'
            if new:
                msg += f' Use {new} instead.'
            logging.warning(msg)
            return method(*args, **kwargs)
        return wrapped
    return wrapper

def validate_timestamps(log_prefix, **kwargs):
    """
    Logs a warning if any provided timestamps are in the past

    Arguments:
        log_prefix (str): Prefix to be prepended to the logged warning(s)
        kwargs (dict): Timestamps to verify
    """
    now = datetime.datetime.now()
    for key, value in kwargs.items():
        if value and dateparser.parse(str(value)) <= now:
            logging.warning(f'{log_prefix} with `{key}` in the past: {value} (now: {now})')

def is_datetime_str(value):
    """
    Checks to see if the string is a valid datetime format

    Arguments:
        value (str): String to test if valid datetime format
    """
    if isinstance(value, str):
        try:
            return datetime.datetime.fromisoformat(value.replace('Z', '+00:00'))
        except ValueError:
            pass
    return False



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for exceptions.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring

from unittest import TestCase

from ..air_sdk import exceptions

class TestAirError(TestCase):
    def test_init(self):
        err = exceptions.AirError('test', 200)
        self.assertEqual(str(err), 'test')
        self.assertEqual(err.status_code, 200)

class TestAirAuthorizationError(TestCase):
    def test_init(self):
        err = exceptions.AirAuthorizationError('test', 200)
        self.assertEqual(err.message, 'test')
        self.assertEqual(err.status_code, 200)
        self.assertIsInstance(err, exceptions.AirError)

    def test_init_default(self):
        err = exceptions.AirAuthorizationError(status_code=200)
        self.assertEqual(err.message, 'An error occurred when authorizing the Air API')
        self.assertEqual(err.status_code, 200)

class TestAirUnexpectedResponse(TestCase):
    def test_init(self):
        err = exceptions.AirUnexpectedResponse('test')
        self.assertEqual(err.message,
                         'Received an unexpected response from the Air API: test')
        self.assertIsNone(err.status_code)
        self.assertIsInstance(err, exceptions.AirError)

    def test_init_status_code(self):
        err = exceptions.AirUnexpectedResponse('test', status_code=200)
        self.assertEqual(err.message,
                         'Received an unexpected response from the Air API (200): test')
        self.assertEqual(err.status_code, 200)

class TestAirForbiddenError(TestCase):
    def test_init(self):
        err = exceptions.AirForbiddenError('test')
        self.assertEqual(err.message, 'test')
        self.assertEqual(err.status_code, 403)
        self.assertIsInstance(err, exceptions.AirError)

    def test_init_default(self):
        err = exceptions.AirForbiddenError()
        self.assertEqual(err.message, 'Received 403 Forbidden. Please call AirApi.authorize().')
        self.assertEqual(err.status_code, 403)

class TestAirObjectDeleted(TestCase):
    def test_init(self):
        err = exceptions.AirObjectDeleted('foo', 'test')
        self.assertEqual(err.message, 'test')
        self.assertIsInstance(err, exceptions.AirError)

    def test_init_default(self):
        err = exceptions.AirObjectDeleted('foo')
        self.assertEqual(err.message,
                         'foo object has been deleted and should no longer be referenced')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Job module
"""

from . import util
from .air_model import AirModel

class Job(AirModel):
    """
    Manage a Job

    ### json
    Returns a JSON string representation of the job

    ### refresh
    Syncs the job with all values returned by the API

    ### update
    Update the job with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _deletable = False

    def __repr__(self):
        if self._deleted or not self.category:
            return super().__repr__()
        return f'<Job {self.category} {self.id}>'

class JobApi:
    """ High-level interface for the Job API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/job/'

    def get(self, job_id, **kwargs):
        """
        Get an existing job

        Arguments:
            job_id (str): Job ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Job`](/docs/job)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.jobs.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Job START 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{job_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Job(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing jobs

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.jobs.list()
        [<Job START c51b49b6-94a7-4c93-950c-e7fa4883591>, <Job STOP 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Job(self, **job) for job in res.json()]



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
SSH Key module
"""

from . import util
from .air_model import AirModel

class SSHKey(AirModel):
    """
    Manage a SSH Key

    ### delete
    Delete the key. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the key

    ### refresh
    Syncs the key with all values returned by the API
    """
    _updatable = False

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<SSHKey {self.name} {self.id}>'

class SSHKeyApi:
    """ High-level interface for the SSHKey API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/sshkey/'

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing keys

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.ssh_keys.list()
        [<SSHKey mykey c51b49b6-94a7-4c93-950c-e7fa4883591>, <SSHKey test_key 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [SSHKey(self, **key) for key in res.json()]

    @util.required_kwargs(['public_key', 'name'])
    def create(self, **kwargs):
        """
        Add a new public key to your account

        Arguments:
            name (str): Descriptive name for the public key
            public_key (str): Public key
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`SSHKey`](/docs/sshkey)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.ssh_keys.create(name='my_pub_key', public_key='<key_string>')
        <SSHKey my_pub_key 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return SSHKey(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Login module
"""

from . import util
from .air_model import AirModel

class Login(AirModel):
    """
    View login information

    ### json
    Returns a JSON string representation of the login info

    ### refresh
    Syncs the login info with all values returned by the API
    """
    _deletable = False
    _updatable = False

    def __repr__(self):
        if self._deleted:
            return super().__repr__()
        return f'<Login {self.id}>'

class LoginApi:
    """ High-level interface for the Login API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/login/'

    def get(self, **kwargs):
        """
        Get login information or start an OAuth request. This is equivalent to `login.list()`.

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Login`](/docs/login)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.login.get()
        <Login>
        ```
        """
        return self.list(**kwargs)

    def list(self, **kwargs):
        """
        Get login information or start an OAuth request. This is equivalent to `login.get()`.

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Login`](/docs/login)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.login.get()
        <Login>
        ```
        """
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res)
        return Login(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for air_api.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,protected-access
#pylint: disable=arguments-differ,unused-argument,no-member,too-many-public-methods

from json import JSONDecodeError
from unittest import TestCase
from unittest.mock import MagicMock, patch
import datetime as dt

import pytest
import requests

from ..air_sdk import air_api
from ..air_sdk.account import AccountApi
from ..air_sdk.air_model import AirModel, LazyLoaded
from ..air_sdk.capacity import CapacityApi
from ..air_sdk.demo import DemoApi
from ..air_sdk.exceptions import AirAuthorizationError, AirForbiddenError, AirUnexpectedResponse
from ..air_sdk.image import ImageApi
from ..air_sdk.interface import InterfaceApi
from ..air_sdk.job import JobApi
from ..air_sdk.link import LinkApi
from ..air_sdk.login import LoginApi
from ..air_sdk.marketplace import MarketplaceApi
from ..air_sdk.node import NodeApi
from ..air_sdk.organization import OrganizationApi
from ..air_sdk.permission import PermissionApi
from ..air_sdk.resource_budget import ResourceBudgetApi
from ..air_sdk.service import ServiceApi
from ..air_sdk.ssh_key import SSHKeyApi
from ..air_sdk.simulation import SimulationApi
from ..air_sdk.simulation_interface import SimulationInterfaceApi
from ..air_sdk.simulation_node import SimulationNodeApi
from ..air_sdk.token import TokenApi
from ..air_sdk.topology import TopologyApi
from ..air_sdk.worker import WorkerApi

class TestAirSession(TestCase):
    def setUp(self):
        self.session = air_api.AirSession()

    def test_init(self):
        self.assertIsInstance(self.session, requests.Session)

    @patch('air_sdk.air_sdk.air_api.requests.Session.rebuild_auth')
    @patch('air_sdk.air_sdk.air_api.urlparse')
    def test_rebuild_auth_allowed(self, mock_parse, mock_rebuild):
        mock_url = MagicMock()
        mock_url.hostname = 'air.nvidia.com'
        mock_parse.return_value = mock_url
        mock_req = MagicMock()
        self.session.rebuild_auth(mock_req, None)
        mock_parse.assert_called_with(mock_req.url)
        mock_rebuild.assert_not_called()

    @patch('air_sdk.air_sdk.air_api.requests.Session.rebuild_auth')
    @patch('air_sdk.air_sdk.air_api.urlparse')
    def test_rebuild_auth_not_allowed(self, mock_parse, mock_rebuild):
        mock_url = MagicMock()
        mock_url.hostname = 'air.evil.com'
        mock_parse.return_value = mock_url
        mock_req = MagicMock()
        mock_res = MagicMock()
        self.session.rebuild_auth(mock_req, mock_res)
        mock_rebuild.assert_called_with(mock_req, mock_res)

class TestAirApi(TestCase):
    @patch('air_sdk.air_sdk.air_api.AirSession')
    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def setUp(self, mock_raise, mock_session):
        self.session = mock_session
        self.req = self.session.return_value
        self.req.headers = {}
        self.api = air_api.AirApi('http://test/api/', 'v1', bearer_token='foo')

    def test_init(self):
        self.assertEqual(self.api.client, self.req)
        self.assertEqual(self.api.client.headers['content-type'], 'application/json')
        self.assertEqual(self.api.api_url, 'http://test/api/v1')
        self.assertEqual(self.api.token, 'foo')
        self.assertIsNone(self.api.username)

    @patch('air_sdk.air_sdk.air_api.AirApi.authorize')
    def test_init_authorize(self, mock_auth):
        self.api = air_api.AirApi('http://test/api/', 'v1', bearer_token='foo')
        mock_auth.assert_called_with(bearer_token='foo')

    def test_accounts(self):
        self.assertIsInstance(self.api.accounts, AccountApi)

    def test_api_tokens(self):
        self.assertIsInstance(self.api.api_tokens, TokenApi)

    def test_capacity(self):
        self.assertIsInstance(self.api.capacity, CapacityApi)

    def test_demos(self):
        self.assertIsInstance(self.api.demos, DemoApi)

    def test_images(self):
        self.assertIsInstance(self.api.images, ImageApi)

    def test_interfaces(self):
        self.assertIsInstance(self.api.interfaces, InterfaceApi)

    def test_jobs(self):
        self.assertIsInstance(self.api.jobs, JobApi)

    def test_links(self):
        self.assertIsInstance(self.api.links, LinkApi)

    def test_login(self):
        self.assertIsInstance(self.api.login, LoginApi)

    def test_marketplace(self):
        self.assertIsInstance(self.api.marketplace, MarketplaceApi)

    def test_node(self):
        self.assertIsInstance(self.api.node, NodeApi)

    def test_nodes(self):
        self.assertIsInstance(self.api.nodes, NodeApi)

    def test_organizations(self):
        self.assertIsInstance(self.api.organizations, OrganizationApi)

    def test_permission(self):
        self.assertIsInstance(self.api.permission, PermissionApi)

    def test_permissions(self):
        self.assertIsInstance(self.api.permissions, PermissionApi)

    def test_resource_budgets(self):
        self.assertIsInstance(self.api.resource_budgets, ResourceBudgetApi)

    def test_service(self):
        self.assertIsInstance(self.api.service, ServiceApi)

    def test_services(self):
        self.assertIsInstance(self.api.services, ServiceApi)

    def test_simulation(self):
        self.assertIsInstance(self.api.simulation, SimulationApi)

    def test_simulations(self):
        self.assertIsInstance(self.api.simulations, SimulationApi)

    def test_simulation_interface(self):
        self.assertIsInstance(self.api.simulation_interface, SimulationInterfaceApi)

    def test_simulation_interfaces(self):
        self.assertIsInstance(self.api.simulation_interfaces, SimulationInterfaceApi)

    def test_simulation_node(self):
        self.assertIsInstance(self.api.simulation_node, SimulationNodeApi)

    def test_simulation_nodes(self):
        self.assertIsInstance(self.api.simulation_nodes, SimulationNodeApi)

    def test_ssh_keys(self):
        self.assertIsInstance(self.api.ssh_keys, SSHKeyApi)

    def test_topology(self):
        self.assertIsInstance(self.api.topology, TopologyApi)

    def test_topologies(self):
        self.assertIsInstance(self.api.topologies, TopologyApi)

    def test_worker(self):
        self.assertIsInstance(self.api.worker, WorkerApi)

    def test_workers(self):
        self.assertIsInstance(self.api.workers, WorkerApi)

    @patch('air_sdk.air_sdk.login.LoginApi.list')
    def test_authorize_token(self, mock_login):
        mock_login.return_value.username = 'john'
        self.api.authorize(bearer_token='foo')
        self.assertEqual(self.api.token, 'foo')
        self.assertEqual(self.api.client.headers['authorization'], 'Bearer foo')
        self.assertEqual(self.api.username, 'john')

    @patch('air_sdk.air_sdk.login.LoginApi.list')
    def test_authorize_password(self, mock_login):
        mock_login.return_value.username = 'john'
        self.api.get_token = MagicMock(return_value='abc123')
        self.api.authorize(username='foo', password='bar')
        self.assertEqual(self.api.token, 'abc123')
        self.assertEqual(self.api.client.headers['authorization'], 'Bearer abc123')
        self.assertEqual(self.api.username, 'john')

    def test_authorize_bad_args(self):
        with pytest.raises(ValueError) as err:
            self.api.authorize()
        self.assertEqual(str(err.value), 'Must include either `bearer_token` or ' + \
                                         '`username` and `password` arguments')

    @patch('air_sdk.air_sdk.air_api.AirApi.post')
    def test_get_token(self, mock_post):
        mock_post.return_value.json.return_value = {'token': 'abc123'}
        res = self.api.get_token('foo', 'bar')
        self.assertEqual(res, 'abc123')
        mock_post.assert_called_with('http://test/api/v1/login/',
                                     json={'username': 'foo', 'password': 'bar'})

    @patch('air_sdk.air_sdk.air_api.AirApi.post')
    def test_get_token_no_token(self, mock_post):
        mock_post.return_value.json.return_value = {'redirect': 'http://test'}
        with self.assertRaises(AirAuthorizationError) as err:
            self.api.get_token('foo', 'bar')
        self.assertEqual(err.exception.message, 'API did not provide a token for foo')

    @patch('air_sdk.air_sdk.air_api.AirApi.post')
    def test_get_token_bad_json(self, mock_post):
        mock_post.return_value.json.side_effect = JSONDecodeError('', '{}', 1)
        with self.assertRaises(AirAuthorizationError) as err:
            self.api.get_token('foo', 'bar')
        self.assertEqual(err.exception.message, 'API did not return a valid JSON response')

    def test_request(self):
        res = self.api._request('GET', 'http://test/', 'test', foo='bar')
        self.api.client.request.assert_called_with('GET', 'http://test/', 'test',
                                                   allow_redirects=False, foo='bar')
        self.assertEqual(res, self.api.client.request.return_value)

    def test_request_403(self):
        self.api.client.request.return_value.status_code = 403
        with self.assertRaises(AirForbiddenError):
            self.api._request('GET', 'http://test/', 'test', foo='bar')

    def test_request_raises(self):
        mock_res = MagicMock()
        self.api.client.request.return_value.raise_for_status = \
            MagicMock(side_effect=requests.exceptions.HTTPError(response=mock_res))
        with self.assertRaises(AirUnexpectedResponse) as err:
            self.api._request('GET', 'http://test/', 'test', foo='bar')
        self.assertEqual(err.exception.message,
                         'Received an unexpected response from the Air API ' + \
                         f'({mock_res.status_code}): {mock_res.text}')
        self.assertEqual(err.exception.status_code, mock_res.status_code)

    @patch('air_sdk.air_sdk.air_api._serialize_dict')
    def test_request_serialized_json_list(self, mock_serialize):
        serialized = ['serialized_foo', 'serialized_bar']
        mock_serialize.side_effect = serialized
        data = ['foo', 'bar']
        self.api._request('GET', 'http://test/', json=data)
        mock_for_assert = MagicMock()
        mock_for_assert(data[0])
        mock_for_assert(data[1])
        self.assertEqual(mock_serialize.mock_calls, mock_for_assert.mock_calls)
        self.api.client.request.assert_called_with('GET', 'http://test/', allow_redirects=False,
                                                   json=[serialized[0], serialized[1]])

    @patch('air_sdk.air_sdk.air_api._serialize_dict')
    def test_request_serialized_json(self, mock_serialize):
        self.api._request('GET', 'http://test/', json='foo')
        mock_serialize.assert_called_with('foo')
        self.api.client.request.assert_called_with('GET', 'http://test/', allow_redirects=False,
                                                   json=mock_serialize.return_value)

    @patch('air_sdk.air_sdk.air_api._serialize_dict')
    def test_request_serialized_params(self, mock_serialize):
        self.api._request('GET', 'http://test/', params='foo')
        mock_serialize.assert_called_with('foo')
        self.api.client.request.assert_called_with('GET', 'http://test/', allow_redirects=False,
                                                   params=mock_serialize.return_value)

    def test_request_redirect(self):
        self.api.client.request.return_value.status_code = 301
        self.api.client.request.return_value.headers = {'Location': 'http://air.nvidia.com/'}
        self.api._request('GET', 'http://test/', json={'foo': 'bar'})
        self.api.client.request.assert_called_with('GET', 'http://air.nvidia.com/',
                                                   json={'foo': 'bar'})
        self.assertEqual(self.api.client.request.call_count, 3)

    def test_request_redirect_ignored(self):
        self.api.client.request.return_value.status_code = 301
        self.api.client.request.return_value.headers = {'Location': 'http://air.evil.com/'}
        self.api._request('GET', 'http://test/', json={'foo': 'bar'})
        self.assertEqual(self.api.client.request.call_count, 2)

    def test_get(self):
        self.api._request = MagicMock()
        self.api.get('http://test/api/v1/foo/', 'arg1', arg2='test')
        self.api._request.assert_called_with('GET', 'http://test/api/v1/foo/', 'arg1', arg2='test')

    def test_post(self):
        self.api._request = MagicMock()
        self.api.post('http://test/api/v1/foo/', 'arg1', arg2='test')
        self.api._request.assert_called_with('POST', 'http://test/api/v1/foo/', 'arg1', arg2='test')

    def test_put(self):
        self.api._request = MagicMock()
        self.api.put('http://test/api/v1/foo/', 'arg1', arg2='test')
        self.api._request.assert_called_with('PUT', 'http://test/api/v1/foo/', 'arg1', arg2='test')

    def test_patch(self):
        self.api._request = MagicMock()
        self.api.patch('http://test/api/v1/foo/', 'arg1', arg2='test')
        self.api._request.assert_called_with('PATCH', 'http://test/api/v1/foo/', 'arg1',
                                             arg2='test')

    def test_delete(self):
        self.api._request = MagicMock()
        self.api.delete('http://test/api/v1/foo/', 'arg1', arg2='test')
        self.api._request.assert_called_with('DELETE', 'http://test/api/v1/foo/', 'arg1',
                                             arg2='test')

class TestHelpers(TestCase):
    def test_normalize_api_version(self):
        res = air_api._normalize_api_version('v1')
        self.assertEqual(res, 'v1')

    def test_normalize_api_version_str(self):
        res = air_api._normalize_api_version('1')
        self.assertEqual(res, 'v1')

    def test_normalize_api_version_int(self):
        res = air_api._normalize_api_version(1)
        self.assertEqual(res, 'v1')

    def test_normalize_api_url(self):
        res = air_api._normalize_api_url('http://localhost/api/')
        self.assertEqual(res, 'http://localhost/api/')

    def test_normalize_api_url_nothing(self):
        res = air_api._normalize_api_url('http://localhost')
        self.assertEqual(res, 'http://localhost/api/')

    def test_normalize_api_url_no_slash(self):
        res = air_api._normalize_api_url('http://localhost/api')
        self.assertEqual(res, 'http://localhost/api/')

    def test_serialize_dict_air_model(self):
        mock_model = AirModel(MagicMock(), id='abc123')
        test_dict = {'test': mock_model}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': 'abc123'})

    def test_serialize_dict_lazy_load(self):
        mock_model = LazyLoaded('abc123', 'test')
        test_dict = {'test': mock_model}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': 'abc123'})

    def test_serialize_dict_dict(self):
        test_dict = {'test': {'foo': 'bar'}}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': {'foo': 'bar'}})

    @patch('air_sdk.air_sdk.air_api._serialize_list')
    def test_serialize_dict_list(self, mock_list):
        test_dict = {'test': ['foo']}
        res = air_api._serialize_dict(test_dict)
        mock_list.assert_called_with(['foo'])
        self.assertDictEqual(res, {'test': mock_list.return_value})

    def test_serialize_dict_datetime(self):
        time = dt.datetime(2030, 12, 12, 22, 5, 3)
        test_dict = {'test': {'foo': time}}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': {'foo': '2030-12-12T22:05:03'}})

    def test_serialize_dict(self):
        test_dict = {'test': 'foo'}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': 'foo'})

    def test_serialize_dict_private(self):
        test_dict = {'test': 'foo', '_private': 'bar'}
        res = air_api._serialize_dict(test_dict)
        self.assertDictEqual(res, {'test': 'foo'})

    def test_serialize_list_air_model(self):
        mock_model = AirModel(MagicMock(), id='abc123')
        test_list = [mock_model]
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, ['abc123'])

    def test_serialize_list_lazy_load(self):
        mock_model = LazyLoaded('abc123', 'test')
        test_list = [mock_model]
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, ['abc123'])

    @patch('air_sdk.air_sdk.air_api._serialize_dict')
    def test_serialize_list_dict(self, mock_dict):
        test_list = [{'foo': 'bar'}]
        mock_dict.called_with({'foo': 'bar'})
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, [mock_dict.return_value])

    def test_serialize_list_list(self):
        test_list = [['foo']]
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, test_list)

    def test_serialize_list(self):
        test_list = ['test']
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, ['test'])

    def test_serialize_list_private(self):
        test_list = ['_private', 'test']
        res = air_api._serialize_list(test_list)
        self.assertListEqual(res, ['test'])



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for simulation_node.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,unused-argument
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import simulation_node

class TestSimulationNode(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.api.url = 'http://testserver/api/'
        self.model = simulation_node.SimulationNode(self.api)
        self.model.id = 'abc123'

    def test_init_(self):
        self.assertFalse(self.model._deletable)
        self.assertTrue(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<SimulationNode {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))


    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_instructions(self, mock_raise):
        res = self.model.create_instructions(executor='shell', data='echo')
        self.api.client.post.assert_called_with(f'{self.api.url}abc123/instructions/',
                                                json={'executor': 'shell', 'data': 'echo'})
        mock_raise.assert_called_with(self.api.client.post.return_value, status_code=201,
                                      data_type=str)
        self.assertEqual(res, {'id': self.api.client.post.return_value.json.return_value})

    def test_create_instructions_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.model.create_instructions(executor='shell')
        self.assertTrue('requires data' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.model.create_instructions(data='foo')
        self.assertTrue('requires executor' in str(err.exception))

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_instructions_list(self, mock_raise):
        self.model.create_instructions(executor='shell', data=['line1', 'line2'])
        self.api.client.post.assert_called_with(f'{self.api.url}abc123/instructions/',
                                                json={'executor': 'shell', 'data': 'line1\nline2'})

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list_instructions(self, mock_raise):
        res = self.model.list_instructions(foo='bar')
        self.api.client.get.assert_called_with(f'{self.api.url}{self.model.id}/instructions/',
                                               params={'foo': 'bar'})
        mock_raise.assert_called_with(self.api.client.get.return_value, data_type=list)
        self.assertEqual(res, self.api.client.get.return_value.json.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_delete_instructions(self, mock_raise):
        self.model.delete_instructions()
        self.api.client.delete.assert_called_with(f'{self.api.url}{self.model.id}/instructions/')
        mock_raise.assert_called_with(self.api.client.delete.return_value, status_code=204,
                                      data_type=None)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_control(self, mock_raise):
        res = self.model.control(action='test')
        self.api.client.post.assert_called_with(f'{self.api.url}{self.model.id}/control/',
                                                json={'action': 'test'})
        mock_raise.assert_called_with(self.api.client.post.return_value)
        self.assertEqual(res, self.api.client.post.return_value.json.return_value)

    def test_control_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.model.control()
        self.assertTrue('requires action' in str(err.exception))

    @patch('air_sdk.air_sdk.simulation_node.SimulationNode.control')
    def test_rebuild(self, mock_control):
        self.model.rebuild(foo='bar')
        mock_control.assert_called_with(action='rebuild', foo='bar')

    @patch('air_sdk.air_sdk.simulation_node.SimulationNode.control')
    def test_reset(self, mock_control):
        self.model.reset(foo='bar')
        mock_control.assert_called_with(action='reset', foo='bar')

class TestSimulationNodeApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = simulation_node.SimulationNodeApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/simulation-node/')

    @patch('air_sdk.air_sdk.simulation_node.SimulationNodeApi.get')
    def test_update_simulation_node(self, mock_get):
        self.api.update_simulation_node('abc123', {'foo': 'bar'})
        mock_get.assert_called_with('abc123')
        mock_get.return_value.update.assert_called_with(foo='bar')

    @patch('air_sdk.air_sdk.simulation_node.SimulationNodeApi.list')
    def test_get_simulation_nodes(self, mock_list):
        res = self.api.get_simulation_nodes(foo='bar')
        mock_list.assert_called_with(foo='bar')
        self.assertEqual(res, mock_list.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/simulation-node/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, simulation_node.SimulationNode)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/simulation-node/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], simulation_node.SimulationNode)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for topology.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,unused-argument
import io
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import topology

class TestTopology(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.api.url = 'http://testserver/api/'
        self.model = topology.Topology(self.api)
        self.model.id = 'abc123'
        self.model.name = 'test'

    def test_init_(self):
        self.assertTrue(self.model._deletable)
        self.assertTrue(self.model._updatable)
        self.assertListEqual(topology.Topology._ignored_update_fields, ['links', 'nodes'])

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Topology {self.model.name} {self.model.id}>')

    def test_add_permission(self):
        res = self.model.add_permission('me@test.com', foo='bar')
        self.api.client.permissions.create.assert_called_with(email='me@test.com',
                                                              topology=self.model.id, foo='bar')
        self.assertEqual(res, self.api.client.permissions.create.return_value)

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))


class TestTopologyApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = topology.TopologyApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/topology/')

    @patch('air_sdk.air_sdk.topology.TopologyApi.list')
    def test_get_topologies(self, mock_list):
        self.assertEqual(self.api.get_topologies(), mock_list.return_value)

    @patch('air_sdk.air_sdk.topology.TopologyApi.create')
    def test_create_topology(self, mock_create):
        res = self.api.create_topology(dot='test')
        mock_create.assert_called_with(dot='test', json=None)
        self.assertEqual(res, mock_create.return_value)

    @patch('air_sdk.air_sdk.topology.TopologyApi.get')
    def test_update_topology(self, mock_get):
        self.api.update_topology('abc123', {'foo': 'bar'})
        mock_get.assert_called_with('abc123')
        mock_get.return_value.update.assert_called_with(foo='bar')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/topology/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, topology.Topology)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/topology/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], topology.Topology)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_json(self, mock_raise):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(json={'foo': 'bar'})
        self.client.post.assert_called_with(f'{self.client.api_url}/topology/',
                                            json={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, topology.Topology)
        self.assertEqual(res.id, 'abc')

    @patch('os.path.isfile', return_value=False)
    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_dot(self, mock_raise, *args):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(dot='test')
        self.client.post.assert_called_with(f'{self.client.api_url}/topology/', data=b'test',
                                            headers={'Content-type': 'text/vnd.graphviz'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, topology.Topology)
        self.assertEqual(res.id, 'abc')

    @patch('os.path.isfile', return_value=False)
    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_dot_file(self, mock_raise, *args):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        mock_file = MagicMock(spec=io.IOBase)
        res = self.api.create(dot=mock_file)
        self.client.post.assert_called_with(f'{self.client.api_url}/topology/', data=mock_file,
                                            headers={'Content-type': 'text/vnd.graphviz'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, topology.Topology)
        self.assertEqual(res.id, 'abc')

    @patch('builtins.open')
    @patch('os.path.isfile', return_value=True)
    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create_dot_file_path(self, mock_raise, mock_isfile, mock_open):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        file_path = '/tmp/topo.dot'
        res = self.api.create(dot=file_path)
        self.client.post.assert_called_with(f'{self.client.api_url}/topology/',
                                            data=mock_open.return_value.read.return_value,
                                            headers={'Content-type': 'text/vnd.graphviz'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, topology.Topology)
        self.assertEqual(res.id, 'abc')
        mock_isfile.assert_called_once_with(file_path)
        mock_open.assert_called_once_with(file_path, 'r')

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create()
        self.assertTrue('requires one of the following: (\'json\', \'dot\')' in str(err.exception))



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for image.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import image, organization

class TestImage(TestCase):
    def setUp(self):
        self.mock_api = MagicMock()
        self.model = image.Image(self.mock_api)
        self.model.id = 'abc123'
        self.model.name = 'ubuntu'
        self.org1 = organization.Organization(self.mock_api, id='xyz456', name='NVIDIA')

    def test_init_(self):
        self.assertTrue(self.model._deletable)
        self.assertTrue(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Image {self.model.name} {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_copy(self, mock_raise):
        new_id = 'def456'
        mock_post = self.mock_api.client.post
        mock_post.return_value.json.return_value = {'id': new_id, 'name': 'new-image'}
        res = self.model.copy(self.org1)
        mock_post.assert_called_once_with(f'{self.mock_api.url}{self.model.id}/copy/',
                                          json={'organization': self.org1})
        mock_raise.assert_called_once_with(self.mock_api.client.post.return_value, status_code=201)
        self.assertEqual(res.id, new_id)

    @patch('builtins.open')
    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_upload(self, mock_raise, mock_open):
        self.model.upload('myfile')
        mock_put = self.mock_api.client.put
        mock_put.assert_called_with(f'{self.mock_api.url}{self.model.id}/upload/',
                                    data=mock_open.return_value.__enter__.return_value)
        mock_open.assert_called_with('myfile', 'rb')
        mock_raise.assert_called_with(mock_put.return_value, status_code=204, data_type=None)

class TestImageApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = image.ImageApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/image/')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/image/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, image.Image)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/image/', params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], image.Image)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create(self, mock_raise):
        org = 'acb123'
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(name='myimage', organization=org)
        self.client.post.assert_called_with(f'{self.client.api_url}/image/',
                                            json={'name': 'myimage', 'organization': org})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, image.Image)
        self.assertEqual(res.id, 'abc')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.image.Image.upload')
    def test_create_upload(self, mock_upload, mock_raise):
        org = 'abc123'
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(name='myimage', filename='myfile', organization=org)
        self.client.post.assert_called_with(f'{self.client.api_url}/image/',
                                            json={'name': 'myimage', 'filename': 'myfile',
                                                  'organization': org})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        mock_upload.assert_called_with('myfile')
        self.assertIsInstance(res, image.Image)

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create(organization='abc123')
        self.assertTrue('requires name' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(name='abc123')
        self.assertTrue('requires organization' in str(err.exception))



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Base classes for AIR object models
"""
from datetime import date, datetime
import json

from . import util
from .exceptions import AirObjectDeleted

class AirModel:
    """ Base class for AIR object models """
    model_keys = {'account': 'accounts', 'base_simulation': 'simulations', 'bios': 'images',
                  'connection': 'links', 'demo': 'demos', 'interface': 'simulation_interfaces',
                  'interfaces': {'Node': 'interfaces', 'SimulationNode': 'simulation_interfaces',
                                 'Link': 'interfaces'},
                  'job': 'jobs', 'last_worker': 'worker',
                  'node': {'Interface': 'nodes', 'NodeInstruction': 'simulation_nodes',
                           'SimulationInterface': 'simulation_nodes',
                           'TopologyInstruction': 'nodes'},
                  'nodes': 'simulation_nodes',
                  'original': {'SimulationInterface': 'interfaces',
                               'SimulationNode': 'nodes'},
                  'organization': 'organizations', 'os': 'images', 'preferred_worker': 'workers',
                  'services': 'services', 'simulation': 'simulations', 'topology': 'topologies',
                  'worker': 'workers'}

    def __init__(self, api, **kwargs):
        self._deleted = False
        super().__setattr__('_updatable', getattr(self, '_updatable', True))
        super().__setattr__('_deletable', getattr(self, '_deletable', True))
        self._api = api
        self._load(**kwargs)

    def _load(self, **kwargs):
        for key, value in kwargs.items():
            _value = value
            datetime_obj = util.is_datetime_str(value)
            if datetime_obj:
                _value = datetime_obj
            if key in self.model_keys and value:
                if isinstance(value, list) and not isinstance(value, LazyLoadedList):
                    _value = LazyLoadedList([LazyLoaded(id=_get_item_id(item),
                                                        model=self._get_model_key(key))
                                             for item in value], self._api)
                elif isinstance(value, (LazyLoaded, LazyLoadedList)):
                    _value = value
                elif value.startswith('http'):
                    _value = LazyLoaded(id=_value.split('/')[6], model=self._get_model_key(key))
                else:
                    _value = LazyLoaded(id=_value, model=self._get_model_key(key))
            super().__setattr__(key, _value)

    def __repr__(self):
        repr_str = super().__repr__()
        if self._deleted:
            repr_str = f'<Deleted Object ({repr_str})>'
        return repr_str

    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if name == '_deleted':
            return value
        if self._deleted:
            raise AirObjectDeleted(type(self))
        if isinstance(value, LazyLoaded):
            value = getattr(self._api.client, value.model).get(value.id)
            super().__setattr__(name, value)
        return value

    def __setattr__(self, name, value):
        if name == '_deleted' or not self._updatable:
            return super().__setattr__(name, value)
        try:
            original = super().__getattribute__(name)
            api = super().__getattribute__('_api')
            id = super().__getattribute__('id') #pylint: disable=redefined-builtin
        except AttributeError:
            original = None
            api = None
            id = None
        if not name.startswith('_') and api and id and original != value:
            self._patch(name, value)
        return super().__setattr__(name, value)

    def _get_model_key(self, key):
        value = self.model_keys[key]
        if isinstance(value, dict):
            value = self.model_keys[key][self.__class__.__name__]
        return value

    def _patch(self, key, value):
        url = f'{self._api.url}{self.id}/'
        res = self._api.client.patch(url, json={key: value})
        util.raise_if_invalid_response(res)

    def update(self, **kwargs):
        """
        Update the object with the provided data

        Arguments:
            kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
        """
        if not self._updatable:
            raise NotImplementedError(f'{self.__class__.__name__} does not support updates')
        url = f'{self._api.url}{self.id}/'
        self.refresh()
        self.__dict__.update(kwargs)
        payload = self.__dict__
        ignored_fields = getattr(self, '_ignored_update_fields', None)
        if ignored_fields:
            allowed_payload = {}
            for key, value in payload.items():
                if key not in ignored_fields:
                    allowed_payload[key] = value
            payload = allowed_payload
        res = self._api.client.put(url, json=payload)
        util.raise_if_invalid_response(res)

    def delete(self):
        """
        Delete the object. Once successful, the object should no longer be used and will raise
        [`AirDeletedObject`](/docs/exceptions) when referenced.

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed
        """
        if not self._deletable:
            raise NotImplementedError(f'{self.__class__.__name__} does not support deletes')
        url = f'{self._api.url}{self.id}/'
        res = self._api.client.delete(url)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)
        self._deleted = True

    def refresh(self):
        """ Syncs the object with all values returned by the API """
        self._load(**self._api.get(self.id).__dict__)

    def json(self):
        """ Returns a JSON string representation of the object """
        payload = {}
        for key, value in self.__dict__.items():
            if isinstance(value, (datetime, date)):
                value = value.isoformat()
            if key.startswith('_'):
                continue
            if isinstance(value, (AirModel, LazyLoaded)):
                payload[key] = value.id
            elif isinstance(value, LazyLoadedList):
                payload[key] = [obj.id for obj in value.__iter__(skip_load=True)]
            else:
                payload[key] = value
        return json.dumps(payload)

class LazyLoaded:
    """ A lazy object whose data will be loaded later """
    def __init__(self, id, model): #pylint: disable=redefined-builtin
        self.id = id
        self.model = model

    def __repr__(self):
        model_str = self.model.capitalize()
        if model_str == 'Topologies':
            model_str = 'Topology'
        elif model_str.endswith('s'):
            model_str = model_str[:-1]
        return f'<air_sdk.air_model.LazyLoaded {model_str} {self.id}>'

class LazyLoadedList(list):
    """ A list whose items are LazyLoaded """
    def __init__(self, items, api):
        self._api = api
        super().__init__(items)

    def __getitem__(self, index):
        value = super().__getitem__(index)
        if isinstance(value, LazyLoaded):
            value = getattr(self._api.client, value.model).get(value.id)
            self[index] = value
        return value

    def __iter__(self, skip_load=False):
        items = super().__iter__()
        for item in items:
            if isinstance(item, LazyLoaded) and not skip_load:
                yield getattr(self._api.client, item.model).get(item.id)
            else:
                yield item

def _get_item_id(item):
    if isinstance(item, dict):
        return item['id']
    try:
        return item.split('/')[6]
    except (AttributeError, IndexError):
        return item



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Worker module
"""

from . import util
from .air_model import AirModel

class Worker(AirModel):
    """
    Manage a Worker

    ### json
    Returns a JSON string representation of the worker

    ### refresh
    Syncs the worker with all values returned by the API

    ### update
    Update the worker with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _deletable = False

    def __repr__(self):
        if self._deleted or not self.fqdn:
            return super().__repr__()
        return f'<Worker {self.fqdn} {self.id}>'

    @util.deprecated('<worker_instance>.available')
    def set_available(self, available):
        """
        Sets a worker's `available` value in AIR

        Arguments:
        available (bool)
        """
        self.available = available #pylint: disable=attribute-defined-outside-init

class WorkerApi:
    """ High-level interface for the Worker API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/worker/'

    @util.deprecated('WorkerApi.list()')
    def get_workers(self, **kwargs): #pylint: disable=missing-function-docstring
        return self.list(**kwargs)

    @util.deprecated('Worker.update()')
    def update_worker(self, worker_id, **kwargs): #pylint: disable=missing-function-docstring
        worker = self.get(worker_id)
        return worker.update(**kwargs)

    def get(self, worker_id, **kwargs):
        """
        Get an existing worker

        Arguments:
            worker_id (str): Worker ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Worker`](/docs/worker)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.workers.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Worker worker01 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{worker_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Worker(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing workers

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.workers.list()
        [<Worker worker01 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Worker worker02 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Worker(self, **worker) for worker in res.json()]

    @util.required_kwargs(['cpu', 'memory', 'storage', 'ip_address', 'port_range', 'username',
                           'password'])
    def create(self, **kwargs):
        #pylint: disable=line-too-long
        """
        Create a new worker

        Arguments:
            cpu (int): Number of vCPUs the worker can support
            memory (int): Amount of memory (in MB) a worker can support
            storage (int): Amount of storage (in GB) a worker can support
            ip_address (str): Internal IP address
            port_range (str): Range of ports available on the worker
            username (str): Worker username for API access
            password (str): Worker password for API access
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Worker`](/docs/worker)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.workers.create(cpu=100, memory=200000, storage=1000, ip_address='10.1.1.1', port_range='10000-30000', username='worker01', password='secret')
        <Worker my_sim 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """ #pylint: enable=line-too-long
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Worker(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for marketplace.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import marketplace

class TestMarketplace(TestCase):
    def setUp(self):
        self.model = marketplace.Marketplace(MagicMock())
        self.model.id = 'abc123'
        self.model.name = 'test'

    def test_init_(self):
        self.assertFalse(self.model._deletable)
        self.assertFalse(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model),
                         f'<Marketplace Demo \'{self.model.name}\' {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

class TestMarketplaceApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = marketplace.MarketplaceApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/marketplace/demo/')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/marketplace/demo/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, marketplace.Marketplace)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/marketplace/demo/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], marketplace.Marketplace)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Simulation module
"""

from . import util
from .air_model import AirModel

class Simulation(AirModel):
    """
    Manage a Simulation

    ### json
    Returns a JSON string representation of the simulation

    ### refresh
    Syncs the simulation with all values returned by the API

    ### update
    Update the simulation with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _deletable = False

    def __repr__(self):
        if self._deleted or not self.title:
            return super().__repr__()
        return f'<Simulation \'{self.title}\' {self.id}>'

    def create_service(self, name, interface, dest_port, **kwargs):
        """
        Create a new service for this simulation

        Arguments:
            name (str): Name of the service
            interface (str | `SimulationInterface`): Interface that the service should be created
                for. This can be provided in one of the following formats:
                - [`SimulationInterface`](/docs/simulation-interface) object
                - ID of a [`SimulationInterface`](/docs/simulation-interface)
                - String in the format of 'node_name:interface_name'
            dest_port (int): Service port number
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Service`](/docs/service)

        Example:
        ```
        >>> simulation.create_service('myservice', 'oob-mgmt-server:eth0', 22, service_type='ssh')
        <Service myservice cc18d746-4cf0-4dd3-80c0-e7df68bbb782>
        >>> simulation.create_service('myservice', simulation_interface, 22, service_type='ssh')
        <Service myservice 9603d0d5-5526-4a0f-91b8-a600010d0091>
        ```
        """
        service = self._api.client.services.create(simulation=self.id, name=name,
                                                   interface=interface, dest_port=dest_port,
                                                   **kwargs)
        self.refresh()
        return service

    def add_permission(self, email, **kwargs):
        """
        Adds permission for a given user to this simulation.

        Arguments:
            email (str): Email address of the user being given permission
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Permission`](/docs/permission)

        Example:
        ```
        >>> simulation.add_permission('mrobertson@nvidia.com', write_ok=True)
        <Permission 217bea68-7048-4262-9bbc-b98ab16c603e>
        ```
        """
        return self._api.client.permissions.create(email=email, simulation=self.id, **kwargs)

    @util.required_kwargs(['action'])
    def control(self, **kwargs):
        """
        Sends a control command to the simulation.

        Arguments:
            action (str): Control command
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Example:
        ```
        >>> simulation.control(action='destroy')
        {'result': 'success'}
        ```
        """
        url = f'{self._api.url}{self.id}/control/'
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res)
        return res.json()

    def load(self):
        """ Alias for `start()` """
        self.start()

    def start(self):
        """ Start/load the simulation """
        self.control(action='load')
        self.refresh()

    def stop(self):
        """ Alias for `store()` """
        self.store()

    def store(self):
        """ Store and power off the simulation """
        self.control(action='store')
        self.refresh()

    def delete(self):
        """ Delete the simulation """
        self.control(action='destroy')
        self._deleted = True

class SimulationApi:
    """ High-level interface for the Simulation API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/simulation/'

    @util.deprecated('SimulationApi.list()')
    def get_simulations(self): #pylint: disable=missing-function-docstring
        return self.list()

    @util.deprecated('SimulationApi.get()')
    def get_simulation(self, simulation_id): #pylint: disable=missing-function-docstring
        return self.get(simulation_id)

    @util.deprecated('SimulationApi.create()')
    def create_simulation(self, **kwargs): #pylint: disable=missing-function-docstring
        return self.create(**kwargs)

    @util.deprecated('Simulation.update()')
    def update_simulation(self, simulation_id, data): #pylint: disable=missing-function-docstring
        sim = self.get(simulation_id)
        sim.update(**data)

    def duplicate(self, simulation, **kwargs):
        """
        Duplicate/clone an existing simulation

        Arguments:
            simulation (str | `Simulation`): Simulation or ID of the snapshot to be duplicated
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        ([`Simulation`](/docs/simulation), dict): Newly created simulation and response JSON

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.duplicate(simulation=simulation)
        <Simulation my_sim 5ff3f0dc-7db8-4938-8257-765c8e48623a>
        ```
        """
        sim = simulation
        if isinstance(sim, str):
            sim = self.get(simulation)
        kwargs['action'] = 'duplicate'
        response = sim.control(**kwargs)
        return Simulation(self, **response['simulation']), response

    @util.deprecated('Simulation.control()')
    def control(self, simulation_id, action, **kwargs): #pylint: disable=missing-function-docstring
        sim = self.get(simulation_id)
        return sim.control(action=action, **kwargs)

    def get_citc_simulation(self):
        """
        Get the active CITC reference simulation

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.get_citc_simulation()
        <Simulation my_sim b9125419-7c6e-41db-bba9-7d647d63943e>
        ```
        """
        url = self.url + 'citc/'
        res = self.client.get(url)
        util.raise_if_invalid_response(res)
        return Simulation(self, **res.json())

    def get(self, simulation_id, **kwargs):
        """
        Get an existing simulation

        Arguments:
            simulation_id (str): Simulation ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Simulation my_sim 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{simulation_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Simulation(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing simulations

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.list()
        [<Simulation sim1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Simulation sim2 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Simulation(self, **simulation) for simulation in res.json()]

    @util.required_kwargs(['topology'])
    def create(self, **kwargs):
        """
        Create a new simulation

        Arguments:
            topology (str | `Topology`): `Topology` or ID
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.create(topology=topology, title='my_sim')
        <Simulation my_sim 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """
        util.validate_timestamps('Simulation created', expires_at=kwargs.get('expires_at'),
                                 sleep_at=kwargs.get('sleep_at'))
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Simulation(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Topology module
"""

import io
import os

from . import util
from .air_model import AirModel

class Topology(AirModel):
    """
    Manage a Topology

    ### delete
    Delete the topology. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the topology

    ### refresh
    Syncs the topology with all values returned by the API

    ### update
    Update the topology with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _ignored_update_fields = ['links', 'nodes']

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Topology {self.name} {self.id}>'

    def add_permission(self, email, **kwargs):
        """
        Adds permission for a given user to this topology.

        Arguments:
            email (str): Email address of the user being given permission
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Permission`](/docs/permission)

        Example:
        ```
        >>> topology.add_permission('mrobertson@nvidia.com', write_ok=True)
        <Permission 217bea68-7048-4262-9bbc-b98ab16c603e>
        ```
        """
        return self._api.client.permissions.create(email=email, topology=self.id, **kwargs)

class TopologyApi:
    """ High-level interface for the Topology API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/topology/'

    @util.deprecated('TopologyApi.list()')
    def get_topologies(self): #pylint: disable=missing-function-docstring
        return self.list()

    @util.deprecated('TopologyApi.create()')
    def create_topology(self, json=None, dot=None): #pylint: disable=missing-function-docstring
        return self.create(json=json, dot=dot)

    @util.deprecated('Topology.update()')
    def update_topology(self, topology_id, data): #pylint: disable=missing-function-docstring
        topology = self.get(topology_id)
        return topology.update(**data)

    def get(self, topology_id, **kwargs):
        """
        Get an existing topology

        Arguments:
            topology_id (str): Topology ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Topology`](/docs/topology)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.topologies.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Topology my_network 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{topology_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Topology(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing topologies

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.topologies.list()
        [<Topology my_network1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Topology my_network2 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Topology(self, **topology) for topology in res.json()]

    @util.required_kwargs([('json', 'dot')])
    def create(self, **kwargs):
        #pylint: disable=line-too-long
        """
        Create a new topology. The caller must provide either `dot` (recommended) or `json`.

        Arguments:
            dot (str | fd, optional): Topology in DOT format. This can be passed as a string
                containing the raw DOT data, a path to the DOT file on your local disk,
                or as a file descriptor for a local file
            json (dict, optional): Topology in JSON format

        Returns:
        [`Topology`](/docs/topology)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.topologies.create(dot='/tmp/my_net.dot')
        <Topology my_net 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        >>> air.topologies.create(dot='graph "my sim" { "server1" [ function="server" os="generic/ubuntu1804"] }')
        <Topology my_net 6256baa8-f54b-4190-85c8-1cc574590080>
        >>> air.topologies.create(dot=open('/tmp/my_net.dot', 'r'))
        <Topology my_net a3d09f12-56ff-4889-8e03-3b714d32c3e5>
        ```
        """
        if kwargs.get('json'):
            res = self.client.post(self.url, json=kwargs['json'])
        else:
            if isinstance(kwargs['dot'], io.IOBase):
                payload = kwargs['dot']
            elif os.path.isfile(kwargs['dot']):
                payload = open(kwargs['dot'], 'r').read()
            else:
                payload = kwargs['dot'].encode('utf-8')
            res = self.client.post(self.url, data=payload,
                                   headers={'Content-type': 'text/vnd.graphviz'})
        util.raise_if_invalid_response(res, status_code=201)
        return Topology(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for service.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,unused-argument
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import service
from ..air_sdk.air_model import AirModel

class TestService(TestCase):
    def setUp(self):
        self.model = service.Service(MagicMock())
        self.model.id = 'abc123'
        self.model.name = 'http'

    def test_init_(self):
        self.assertTrue(self.model._deletable)
        self.assertTrue(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Service {self.model.name} {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))


class TestServiceApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = service.ServiceApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/service/')

    @patch('air_sdk.air_sdk.service.ServiceApi.list')
    def test_get_services(self, mock_list):
        self.assertEqual(self.api.get_services(), mock_list.return_value)

    @patch('air_sdk.air_sdk.service.ServiceApi.get')
    def test_get_service(self, mock_get):
        res = self.api.get_service('abc123')
        mock_get.assert_called_with('abc123')
        self.assertEqual(res, mock_get.return_value)

    @patch('air_sdk.air_sdk.service.ServiceApi.create')
    def test_create_service(self, mock_create):
        res = self.api.create_service('abc123', 'test', 'intf', 22, foo='bar')
        mock_create.assert_called_with(simulation='abc123', name='test', interface='intf',
                                       dest_port=22, foo='bar')
        self.assertEqual(res, mock_create.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/service/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, service.Service)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/service/', params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], service.Service)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.service.ServiceApi._resolve_interface')
    def test_create(self, mock_resolve, mock_raise):
        sim = AirModel(MagicMock(), id='xyz123')
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(name='abc123', simulation=sim, interface='test:eth0')
        mock_resolve.assert_called_with('test:eth0', sim)
        self.client.post.assert_called_with(f'{self.client.api_url}/service/',
                                            json={'name': 'abc123', 'simulation': sim,
                                                  'interface': mock_resolve.return_value})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, service.Service)
        self.assertEqual(res.id, 'abc')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.service.ServiceApi._resolve_interface')
    def test_create_id(self, mock_resolve, mock_raise):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        self.api.create(name='abc123', simulation='xyz123', interface='test123')
        mock_resolve.assert_not_called()
        self.client.post.assert_called_with(f'{self.client.api_url}/service/',
                                            json={'name': 'abc123', 'simulation': 'xyz123',
                                                  'interface': 'test123'})

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create(simulation='abc123', interface='xyz123')
        self.assertTrue('requires name' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(name='test', interface='xyz123')
        self.assertTrue('requires simulation' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(name='test', simulation='xyz123')
        self.assertTrue('requires interface' in str(err.exception))

    def test_resolve_interface(self):
        intf1 = MagicMock()
        intf1.name = 'eth0'
        intf2 = MagicMock()
        intf2.name = 'eth1'
        node1 = MagicMock()
        node1.name = 'server'
        node1.interfaces = [intf1, intf2]
        node2 = MagicMock()
        node2.name = 'foo'
        node2.interfaces = [intf1, intf2]
        self.client.nodes.list.return_value = [node1, node2]
        mock_simint = MagicMock()
        self.client.simulation_interfaces.list.return_value = [mock_simint]
        res = self.api._resolve_interface('server:eth0', 'abc123')
        self.client.nodes.list.assert_called_with(simulation='abc123')
        self.client.simulation_interfaces.list.assert_called_with(original=intf1,
                                                                  simulation='abc123')
        self.assertEqual(res, mock_simint)

    def test_resolve_interface_bad_input(self):
        with self.assertRaises(ValueError) as err:
            self.api._resolve_interface('eth0', 'abc123')
        self.assertEqual(str(err.exception),
                         '`interface` must be an Interface object or in the format of ' + \
                         '"node_name:interface_name"')

    def test_resolve_interface_not_found(self):
        with self.assertRaises(ValueError) as err:
            self.api._resolve_interface('server:eth0', 'abc123')
        self.assertEqual(str(err.exception), 'Interface server:eth0 does not exist')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for simulation.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,unused-argument
import datetime as dt
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import simulation

class TestSimulation(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.api.url = 'http://testserver/api/'
        self.model = simulation.Simulation(self.api)
        self.model.id = 'abc123'
        self.model.title = 'test'

    def test_init_(self):
        self.assertFalse(self.model._deletable)
        self.assertTrue(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Simulation \'{self.model.title}\' {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

    @patch('air_sdk.air_sdk.simulation.Simulation.refresh')
    def test_create_service(self, mock_refresh):
        res = self.model.create_service('test', 'intf', 22, foo='bar')
        self.api.client.services.create.assert_called_with(simulation=self.model.id, name='test',
                                                           interface='intf', dest_port=22,
                                                           foo='bar')
        mock_refresh.asserrt_called()
        self.assertEqual(res, self.api.client.services.create.return_value)

    def test_add_permission(self):
        res = self.model.add_permission('me@test.com', foo='bar')
        self.api.client.permissions.create.assert_called_with(email='me@test.com',
                                                              simulation=self.model.id, foo='bar')
        self.assertEqual(res, self.api.client.permissions.create.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_control(self, mock_raise):
        res = self.model.control(action='test')
        self.api.client.post(f'{self.api.url}{self.model.id}/control/', json={'action': 'test'})
        mock_raise.assert_called_with(self.api.client.post.return_value)
        self.assertEqual(res, self.api.client.post.return_value.json.return_value)

    def test_control_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.model.control()
        self.assertTrue('requires action' in str(err.exception))

    @patch('air_sdk.air_sdk.simulation.Simulation.start')
    def test_load(self, mock_start):
        self.model.load()
        mock_start.assert_called()

    @patch('air_sdk.air_sdk.simulation.Simulation.control')
    def test_start(self, mock_control):
        self.model.start()
        mock_control.assert_called_with(action='load')

    @patch('air_sdk.air_sdk.simulation.Simulation.store')
    def test_stop(self, mock_store):
        self.model.stop()
        mock_store.assert_called()

    @patch('air_sdk.air_sdk.simulation.Simulation.control')
    def test_store(self, mock_control):
        self.model.store()
        mock_control.assert_called_with(action='store')

    @patch('air_sdk.air_sdk.simulation.Simulation.control')
    def test_delete(self, mock_control):
        self.model.delete()
        mock_control.assert_called_with(action='destroy')

class TestSimulationApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = simulation.SimulationApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/simulation/')

    @patch('air_sdk.air_sdk.simulation.SimulationApi.list')
    def test_get_simulations(self, mock_list):
        self.assertEqual(self.api.get_simulations(), mock_list.return_value)

    @patch('air_sdk.air_sdk.simulation.SimulationApi.get')
    def test_get_simulation(self, mock_get):
        res = self.api.get_simulation('abc123')
        mock_get.assert_called_with('abc123')
        self.assertEqual(res, mock_get.return_value)

    @patch('air_sdk.air_sdk.simulation.SimulationApi.create')
    def test_create_simulation(self, mock_create):
        res = self.api.create_simulation(foo='bar')
        mock_create.assert_called_with(foo='bar')
        self.assertEqual(res, mock_create.return_value)

    @patch('air_sdk.air_sdk.simulation.SimulationApi.get')
    def test_update_simulation(self, mock_get):
        self.api.update_simulation('abc123', {'foo': 'bar'})
        mock_get.assert_called_with('abc123')
        mock_get.return_value.update.assert_called_with(foo='bar')

    @patch('air_sdk.air_sdk.simulation.SimulationApi.get')
    def test_duplicate(self, mock_get):
        mock_get.return_value.control.return_value = {'simulation': {'test': 'xyz'}}
        sim, res = self.api.duplicate('abc123', foo='bar')
        mock_get.assert_called_with('abc123')
        mock_get.return_value.control.assert_called_with(foo='bar', action='duplicate')
        self.assertIsInstance(sim, simulation.Simulation)
        self.assertEqual(sim.test, 'xyz')
        self.assertEqual(res, mock_get.return_value.control.return_value)

    @patch('air_sdk.air_sdk.simulation.SimulationApi.get')
    def test_duplicate_object(self, mock_get):
        mock_snap = MagicMock()
        mock_snap.control.return_value = {'simulation': {'test': 'xyz'}}
        sim, res = self.api.duplicate(mock_snap, foo='bar')
        mock_get.assert_not_called()
        mock_snap.control.assert_called_with(foo='bar', action='duplicate')
        self.assertIsInstance(sim, simulation.Simulation)
        self.assertEqual(res, mock_snap.control.return_value)

    @patch('air_sdk.air_sdk.simulation.SimulationApi.get')
    def test_control(self, mock_get):
        res = self.api.control('abc123', 'test', foo='bar')
        mock_get.assert_called_with('abc123')
        mock_get.return_value.control.assert_called_with(action='test', foo='bar')
        self.assertEqual(res, mock_get.return_value.control.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get_citc_simulation(self, mock_raise):
        self.client.get.return_value.json.return_value = {'foo': 'bar'}
        res = self.api.get_citc_simulation()
        self.client.get.assert_called_with(f'{self.client.api_url}/simulation/citc/')
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, simulation.Simulation)
        self.assertEqual(res.foo, 'bar')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/simulation/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, simulation.Simulation)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/simulation/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], simulation.Simulation)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.util.validate_timestamps')
    def test_create(self, mock_validate, mock_raise):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(topology='abc123')
        self.client.post.assert_called_with(f'{self.client.api_url}/simulation/',
                                            json={'topology': 'abc123'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, simulation.Simulation)
        self.assertEqual(res.id, 'abc')
        mock_validate.assert_called_with('Simulation created', expires_at=None, sleep_at=None)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.util.validate_timestamps')
    def test_create_timestamps(self, mock_validate, mock_raise):
        self.api.create(topology='abc123', expires_at='expired', sleep_at='sleepy')
        mock_validate.assert_called_with('Simulation created', expires_at='expired',
                                         sleep_at='sleepy')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    @patch('air_sdk.air_sdk.util.validate_timestamps')
    def test_create_datetime(self, mock_validate, mock_raise):
        time = dt.datetime(2030, 12, 12, 22, 5, 3)
        self.api.create(topology='abc123', expires_at=time, sleep_at=time)
        mock_validate.assert_called_with('Simulation created', expires_at=time,
                                         sleep_at=time)

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create()
        self.assertTrue('requires topology' in str(err.exception))



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for worker.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import worker

class TestWorker(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.model = worker.Worker(self.api)
        self.model.fqdn = 'test.test'
        self.model.id = 'abc123'

    def test_init_(self):
        self.assertFalse(self.model._deletable)
        self.assertTrue(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Worker {self.model.fqdn} {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

    @patch('air_sdk.air_sdk.air_model.AirModel._patch')
    def test_set_available(self, mock_patch):
        self.model.available = False
        self.model.set_available(True)
        self.assertTrue(self.model.available)
        mock_patch.assert_called()

class TestWorkerApi(TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.client.api_url = 'http://testserver/api'
        self.api = worker.WorkerApi(self.client)

    def test_init_(self):
        self.assertEqual(self.api.client, self.client)
        self.assertEqual(self.api.url, 'http://testserver/api/worker/')

    @patch('air_sdk.air_sdk.worker.WorkerApi.list')
    def test_get_workers(self, mock_list):
        res = self.api.get_workers(foo='bar')
        mock_list.assert_called_with(foo='bar')
        self.assertEqual(res, mock_list.return_value)

    @patch('air_sdk.air_sdk.worker.WorkerApi.get')
    def test_update_worker(self, mock_get):
        res = self.api.update_worker('abc123', foo='bar')
        mock_get.assert_called_with('abc123')
        mock_get.return_value.update.assert_called_with(foo='bar')
        self.assertEqual(res, mock_get.return_value.update.return_value)

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_get(self, mock_raise):
        self.client.get.return_value.json.return_value = {'test': 'success'}
        res = self.api.get('abc123', foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/worker/abc123/',
                                           params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value)
        self.assertIsInstance(res, worker.Worker)
        self.assertEqual(res.test, 'success')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_list(self, mock_raise):
        self.client.get.return_value.json.return_value = [{'id': 'abc'}, {'id': 'xyz'}]
        res = self.api.list(foo='bar')
        self.client.get.assert_called_with(f'{self.client.api_url}/worker/', params={'foo': 'bar'})
        mock_raise.assert_called_with(self.client.get.return_value, data_type=list)
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], worker.Worker)
        self.assertEqual(res[0].id, 'abc')
        self.assertEqual(res[1].id, 'xyz')

    @patch('air_sdk.air_sdk.util.raise_if_invalid_response')
    def test_create(self, mock_raise):
        self.client.post.return_value.json.return_value = {'id': 'abc'}
        res = self.api.create(cpu=1, memory=2, storage=3, ip_address='10.1.1.1', port_range='1-2',
                              username='foo', password='bar')
        self.client.post.assert_called_with(f'{self.client.api_url}/worker/',
                                            json={'cpu': 1, 'memory': 2, 'storage': 3,
                                                  'ip_address': '10.1.1.1', 'port_range': '1-2',
                                                  'username': 'foo', 'password': 'bar'})
        mock_raise.assert_called_with(self.client.post.return_value, status_code=201)
        self.assertIsInstance(res, worker.Worker)
        self.assertEqual(res.id, 'abc')

    def test_create_required_kwargs(self):
        with self.assertRaises(AttributeError) as err:
            self.api.create(memory=2, storage=3, ip_address='10.1.1.1', port_range='1-2',
                            username='foo', password='bar')
        self.assertTrue('requires cpu' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, storage=3, ip_address='10.1.1.1', port_range='1-2',
                            username='foo', password='bar')
        self.assertTrue('requires memory' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, memory=2, ip_address='10.1.1.1', port_range='1-2',
                            username='foo', password='bar')
        self.assertTrue('requires storage' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, memory=2, storage=3, port_range='1-2', username='foo',
                            password='bar')
        self.assertTrue('requires ip_address' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, memory=2, storage=3, ip_address='10.1.1.1', username='foo',
                            password='bar')
        self.assertTrue('requires port_range' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, memory=2, storage=3, ip_address='10.1.1.1', port_range='1-2',
                            password='bar')
        self.assertTrue('requires username' in str(err.exception))
        with self.assertRaises(AttributeError) as err:
            self.api.create(cpu=1, memory=2, storage=3, ip_address='10.1.1.1', port_range='1-2',
                            username='foo')
        self.assertTrue('requires password' in str(err.exception))



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

""" Exposes the AIR API client module """
from .air_api import *



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
SimulationNode module
"""

from . import util
from .air_model import AirModel

class SimulationNode(AirModel):
    """
    Manage a SimulationNode

    ### json
    Returns a JSON string representation of the simulation node

    ### refresh
    Syncs the simulation node with all values returned by the API

    ### update
    Update the simulation node with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """
    _deletable = False

    def __repr__(self):
        if self._deleted:
            return super().__repr__()
        return f'<SimulationNode {self.id}>'

    @util.required_kwargs(['executor', 'data'])
    def create_instructions(self, **kwargs):
        """
        Create instructions for the `SimulationNode`'s agent to execute

        Arguments:
            data (str | list): Instruction data
            executor (str): Agent executor type
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> simulation_node.create_instructions(data='echo foo', executor='shell')
        {'id': '67f73552-ffdf-4e5f-9881-aeae227604a3'}
        ```
        """
        url = f'{self._api.url}{self.id}/instructions/'
        if isinstance(kwargs['data'], list):
            kwargs['data'] = '\n'.join(kwargs['data'])
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201, data_type=str)
        return {'id': res.json()}

    def list_instructions(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List all instructions for a `SimulationNode`

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> simulation_node.instructions.list()
        [{'id': '56abc69b-489f-429a-aed9-600f26afc956'}, {'id': '7c9c3449-f071-4bbc-bb42-bef04e44d74e'}]
        ```
        """ #pylint: enable=line-too-long
        url = f'{self._api.url}{self.id}/instructions/'
        res = self._api.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return res.json()

    def delete_instructions(self):
        """
        Delete all instructions for a `SimulationNode`

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Instruction delete failed

        Example:
        ```
        >>> simulation_node.instructions.delete()
        ```
        """
        url = f'{self._api.url}{self.id}/instructions/'
        res = self._api.client.delete(url)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    @util.required_kwargs(['action'])
    def control(self, **kwargs):
        """
        Sends a control command to the `SimulationNode`.

        Arguments:
            action (str): Control command
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Example:
        ```
        >>> simulation_node.control(action='reset')
        {'result': 'success'}
        ```
        """
        url = f'{self._api.url}{self.id}/control/'
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res)
        return res.json()

    def rebuild(self, **kwargs):
        """
        Rebuild the `SimulationNode` back to it's initial state. **All existing data will be lost!**
        """
        self.control(action='rebuild', **kwargs)

    def reset(self, **kwargs):
        """ Reset the `SimulationNode` """
        self.control(action='reset', **kwargs)

class SimulationNodeApi:
    """ Wrapper for the SimulationNode API """
    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/simulation-node/'

    @util.deprecated('SimulationNode.update()')
    def update_simulation_node(self, simulation_node_id, data): #pylint: disable=missing-function-docstring
        node = self.get(simulation_node_id)
        node.update(**data)

    @util.deprecated('SimulationNodeApi.list()')
    def get_simulation_nodes(self, **kwargs): #pylint: disable=missing-function-docstring
        return self.list(**kwargs)

    def get(self, simulation_node_id, **kwargs):
        """
        Get an existing simulation node

        Arguments:
            simulation_node_id (str): SimulationNode ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`SimulationNode`](/docs/simulationnode)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulation_nodes.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <SimulationNode my_sim 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{simulation_node_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return SimulationNode(self, **res.json())

    def list(self, **kwargs):
        #pylint: disable=line-too-long
        """
        List existing simulation nodes

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulation_nodes.list()
        [<SimulationNode sim1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <SimulationNode sim2 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """ #pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [SimulationNode(self, **simulation_node) for simulation_node in res.json()]



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for util.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,no-self-use,unused-argument

import datetime

from json import JSONDecodeError
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import exceptions, util

class TestUtil(TestCase):
    def test_raise_if_invalid_response(self):
        mock_res = MagicMock()
        mock_res.status_code = 200
        mock_res.json.return_value = {'id': 'abc123'}
        util.raise_if_invalid_response(mock_res)

    def test_raise_if_invalid_response_status(self):
        mock_res = MagicMock()
        mock_res.status_code = 400
        with self.assertRaises(exceptions.AirUnexpectedResponse) as err:
            util.raise_if_invalid_response(mock_res)
        self.assertEqual(err.exception.message,
                         'Received an unexpected response from the Air API (400): ' +
                         str(mock_res.text))
        self.assertEqual(err.exception.status_code, 400)

    def test_raise_if_invalid_response_no_data(self):
        mock_res = MagicMock()
        mock_res.status_code = 200
        util.raise_if_invalid_response(mock_res, data_type=None)

    def test_raise_if_invalid_bad_json(self):
        mock_res = MagicMock()
        mock_res.status_code = 200
        mock_res.json.side_effect = JSONDecodeError('test', 'foo', 0)
        with self.assertRaises(exceptions.AirUnexpectedResponse) as err:
            util.raise_if_invalid_response(mock_res)
        self.assertEqual(err.exception.message,
                         'Received an unexpected response from the Air API (200): ' +
                         str(mock_res.text))
        self.assertEqual(err.exception.status_code, 200)

    def test_raise_if_invalid_bad_data_type(self):
        mock_res = MagicMock()
        mock_res.status_code = 200
        mock_res.json.return_value = {}
        with self.assertRaises(exceptions.AirUnexpectedResponse) as err:
            util.raise_if_invalid_response(mock_res, data_type=list)
        self.assertEqual(err.exception.message,
                         'Received an unexpected response from the Air API (200): ' + \
                         'Expected API response to be of type <class \'list\'>, ' + \
                         'got <class \'dict\'>')
        self.assertEqual(err.exception.status_code, 200)

    def test_required_kwargs(self):
        @util.required_kwargs(['foo', 'bar'])
        def decorated(**kwargs):
            pass
        with self.assertRaises(AttributeError) as err:
            decorated(foo='test')
        self.assertTrue('requires bar' in str(err.exception))

    def test_required_kwargs_str(self):
        @util.required_kwargs('foo')
        def decorated(**kwargs):
            pass
        with self.assertRaises(AttributeError) as err:
            decorated(f='test')
        self.assertTrue('requires foo' in str(err.exception))

    @patch('logging.warning')
    def test_deprecated(self, mock_log):
        @util.deprecated()
        def decorated():
            pass
        decorated()
        self.assertTrue('has been deprecated and will be removed in a future release.' \
                        in mock_log.call_args[0][0])

    @patch('logging.warning')
    def test_deprecated_new(self, mock_log):
        @util.deprecated('new_func')
        def decorated():
            pass
        decorated()
        self.assertTrue('has been deprecated and will be removed in a future release. ' + \
                        'Use new_func instead.' in mock_log.call_args[0][0])

    @patch('logging.warning')
    def test_validate_timestamps(self, mock_log):
        now = datetime.datetime.now()
        past = now - datetime.timedelta(hours=8)
        util.validate_timestamps('Simulation created', expires_at=None, sleep_at=past)
        log = mock_log.call_args[0][0]
        self.assertTrue(f'Simulation created with `sleep_at` in the past: {past}' in log)

    @patch('logging.warning')
    def test_validate_timestamps_future(self, mock_log):
        now = datetime.datetime.now()
        past = now - datetime.timedelta(hours=8)
        future = now + datetime.timedelta(hours=8)
        util.validate_timestamps('Simulation created', expires_at=past, sleep_at=future)
        log = mock_log.call_args[0][0]
        self.assertTrue(f'Simulation created with `expires_at` in the past: {past}' in log)



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for account.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring
from unittest import TestCase
from unittest.mock import MagicMock

from ..air_sdk import account

class TestAccount(TestCase):
    def setUp(self):
        self.model = account.Account(MagicMock())
        self.model.username = 'foo'
        self.model.id = 'abc123'
        self.assertFalse(self.model._deletable)
        self.assertFalse(self.model._updatable)

    def test_init(self):
        self.assertFalse(self.model._deletable)
        self.assertFalse(self.model._updatable)

    def test_repr(self):
        self.assertEqual(str(self.model), f'<Account {self.model.username} {self.model.id}>')

    def test_repr_deleted(self):
        self.model._deleted = True
        self.assertTrue('Deleted Object' in str(self.model))

class TestAccountApi(TestCase):
    def setUp(self):
        self.mock_client = MagicMock()
        self.mock_client.api_url = 'http://testserver/api'
        self.mock_client.get.return_value.status_code = 200
        self.url = self.mock_client.api_url + '/account/'
        self.api = account.AccountApi(self.mock_client)

    def test_init(self):
        self.assertEqual(self.api.client, self.mock_client)
        self.assertEqual(self.api.url, self.url)

    def test_get(self):
        self.mock_client.get.return_value.json.return_value = {'foo': 'bar'}
        res = self.api.get('abc123', foo='bar')
        self.mock_client.get.assert_called_with(self.url + 'abc123/', params={'foo': 'bar'})
        self.assertIsInstance(res, account.Account)
        self.assertEqual(res.foo, 'bar')

    def test_list(self):
        self.mock_client.get.return_value.json.return_value = [{'foo': 'bar'}]
        res = self.api.list(foo='bar')
        self.mock_client.get.assert_called_with(self.url, params={'foo': 'bar'})
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0], account.Account)
        self.assertEqual(res[0].foo, 'bar')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Tests for air_model.py
"""
#pylint: disable=missing-function-docstring,missing-class-docstring,unused-argument
#pylint: disable=too-many-public-methods,duplicate-code,protected-access
import datetime as dt
from datetime import date, datetime
from unittest import TestCase
from unittest.mock import MagicMock, patch

from ..air_sdk import air_model
from ..air_sdk.exceptions import AirObjectDeleted
from ..air_sdk.node import Node

@patch('air_sdk.air_sdk.util.raise_if_invalid_response')
class TestAirModel(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.model = air_model.AirModel(self.api, foo='bar')
        self.model._api.client.patch = MagicMock()
        self.model._api.client.put = MagicMock()
        self.model._api.client.delete = MagicMock()
        self.model.id = 'abc123'

    def test_init(self, mock_raise):
        self.assertFalse(self.model._deleted)
        self.assertTrue(self.model._updatable)
        self.assertTrue(self.model._deletable)
        self.assertEqual(self.model._api, self.api)
        self.assertEqual(self.model.foo, 'bar')

    def test_model_keys(self, mock_raise):
        model_keys = {'account': 'accounts', 'base_simulation': 'simulations', 'bios': 'images',
                      'connection': 'links', 'demo': 'demos', 'interface': 'simulation_interfaces',
                      'interfaces': {'Node': 'interfaces',
                                     'SimulationNode': 'simulation_interfaces',
                                     'Link': 'interfaces'},
                      'job': 'jobs', 'last_worker': 'worker',
                      'node': {'Interface': 'nodes', 'NodeInstruction': 'simulation_nodes',
                               'SimulationInterface': 'simulation_nodes',
                               'TopologyInstruction': 'nodes'},
                      'nodes': 'simulation_nodes',
                      'original': {'SimulationInterface': 'interfaces',
                                   'SimulationNode': 'nodes'},
                      'organization': 'organizations', 'os': 'images',
                      'preferred_worker': 'workers', 'services': 'services',
                      'simulation': 'simulations', 'topology': 'topologies', 'worker': 'workers'}
        self.assertDictEqual(self.model.model_keys, model_keys)

    def test_load(self, mock_raise):
        model = air_model.AirModel(self.api, normal='http://testserver/api/v1/thing3/abc456')
        self.assertEqual(model.normal, 'http://testserver/api/v1/thing3/abc456')

    def test_load_datetime(self, mock_raise):
        model = air_model.AirModel(self.api, expires_at='2030-12-12T22:05:03')
        self.assertIsInstance(model.expires_at, (datetime, date))

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys', {'lazy_list': 'lazy_api'})
    @patch('air_sdk.air_sdk.air_model.LazyLoadedList.__getitem__')
    def test_load_list(self, mock_list_get, mock_raise):
        model = air_model.AirModel(self.api, lazy_list=['http://testserver/api/v1/thing/abc123'])
        self.assertIsInstance(model.lazy_list, air_model.LazyLoadedList)
        self.assertEqual(model.lazy_list[0].id, mock_list_get.return_value.id)

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys', {'lazy_lazy': 'lazy_api'})
    def test_load_lazy_exists(self, mock_raise):
        lazy = air_model.LazyLoaded('def123', 'lazy')
        model = air_model.AirModel(self.api, lazy_lazy=lazy)
        self.assertEqual(model.lazy_lazy, self.api.client.lazy.get.return_value)

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys', {'lazy_item': 'lazy_api'})
    def test_load_http(self, mock_raise):
        model = air_model.AirModel(self.api, lazy_item='http://testserver/api/v1/thing2/xyz123')
        self.assertEqual(model.lazy_item.id, self.api.client.lazy_api.get.return_value.id)

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys', {'lazy_item': 'lazy_api'})
    def test_load_lazy(self, mock_raise):
        model = air_model.AirModel(self.api, lazy_item='xyz123')
        self.assertEqual(model.lazy_item.id, self.api.client.lazy_api.get.return_value.id)

    def test_repr(self, mock_raise):
        self.assertRegex(str(self.model),
                         r'<air_sdk.air_sdk.air_model.AirModel object at 0x[0-9a-f]+>')

    def test_repr_deleted(self, mock_raise):
        self.model._deleted = True
        self.assertRegex(str(self.model),
                         r'<Deleted Object \(<air_sdk.air_sdk.air_model.AirModel ' +
                         r'object at 0x[0-9a-f]+>\)>')

    def test_getattribute_get_deleted(self, mock_raise):
        self.assertFalse(self.model._deleted)
        self.api.client.get.assert_not_called()

    def test_getattribute_deleted(self, mock_raise):
        self.model._deleted = True
        with self.assertRaises(AirObjectDeleted) as err:
            _ = self.model.foo
        self.assertEqual(err.exception.message,
                         '<class \'air_sdk.air_sdk.air_model.AirModel\'> object has ' + \
                         'been deleted and should no longer be referenced')

    def test_getattribute_lazy(self, mock_raise):
        self.model.lazy = air_model.LazyLoaded('abc123', 'thing')
        self.assertEqual(self.model.lazy, self.api.client.thing.get.return_value)
        self.api.client.thing.get.assert_called_with('abc123')

    def test_setattr_set_deleted(self, mock_raise):
        self.model._deleted = True
        self.assertTrue(self.model._deleted)

    @patch('air_sdk.air_sdk.air_model.AirModel._patch')
    def test_setattr_not_updatable(self, mock_patch, mock_raise):
        self.model._updatable = False
        mock_patch.reset_mock()
        self.model.foo = 'test'
        self.assertEqual(self.model.foo, 'test')
        mock_patch.assert_not_called()

    def test_setattr(self, mock_raise):
        self.model._patch = MagicMock()
        self.model.foo = 'test'
        self.model._patch.assert_called_with('foo', 'test')
        self.assertEqual(self.model.foo, 'test')

    def test_setattr_no_change(self, mock_raise):
        self.model._patch = MagicMock()
        self.model.foo = 'bar'
        self.model._patch.assert_not_called()
        self.assertEqual(self.model.foo, 'bar')

    def test_setattr_internal(self, mock_raise):
        self.model._patch = MagicMock()
        self.model._foo = 'bar'
        self.model._patch.assert_not_called()
        self.assertEqual(self.model._foo, 'bar')

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys', {'simulation': 'simulations'})
    def test_get_model_key(self, mock_raise):
        self.assertEqual(self.model._get_model_key('simulation'), 'simulations')

    @patch('air_sdk.air_sdk.air_model.AirModel.model_keys',
           {'simulation': {'Node': 'a', 'Node2': 'b'}})
    def test_get_model_key_dict(self, mock_raise):
        node = Node(MagicMock())
        self.assertEqual(node._get_model_key('simulation'), 'a')

    def test_patch(self, mock_raise):
        self.model.id = 'abc123'
        self.model._patch('foo', 'bar')
        self.model._api.client.patch.assert_called_with(f'{self.api.url}abc123/',
                                                        json={'foo': 'bar'})
        mock_raise.assert_called_with(self.model._api.client.patch.return_value)

    @patch('air_sdk.air_sdk.air_model.AirModel.refresh')
    def test_update(self, mock_refresh, mock_raise):
        self.model.update(test='new')
        mock_refresh.assert_called()
        self.model._api.client.put.assert_called_with(f'{self.model._api.url}{self.model.id}/',
                                                      json=self.model.__dict__)
        mock_raise.assert_called_with(self.model._api.client.put.return_value)
        self.assertEqual(self.model.test, 'new')

    def test_update_not_updatable(self, mock_raise):
        self.model._updatable = False
        with self.assertRaises(NotImplementedError) as err:
            self.model.update(test='new')
        self.assertEqual(str(err.exception), 'AirModel does not support updates')

    def test_update_ignored_fields(self, mock_raise):
        self.model.ignore_me = True
        self.model._ignored_update_fields = ['ignore_me']
        self.model.update(test='new')
        payload = self.model.__dict__
        del payload['ignore_me']
        self.model._api.client.put.assert_called_with(f'{self.model._api.url}{self.model.id}/',
                                                      json=payload)
        mock_raise.assert_called_with(self.model._api.client.put.return_value)
        self.assertEqual(self.model.test, 'new')

    def test_delete(self, mock_raise):
        self.model.delete()
        self.assertTrue(self.model._deleted)
        self.model._deleted = False
        self.model._api.client.delete.assert_called_with(f'{self.model._api.url}{self.model.id}/')
        mock_raise.assert_called_with(self.model._api.client.delete.return_value, data_type=None,
                                      status_code=204)

    def test_delete_not_deletable(self, mock_raise):
        self.model._deletable = False
        with self.assertRaises(NotImplementedError) as err:
            self.model.delete()
        self.assertEqual(str(err.exception), 'AirModel does not support deletes')

    @patch('air_sdk.air_sdk.AirModel._load')
    def test_refresh(self, mock_load, mock_raise):
        self.model.refresh()
        self.model._api.get.assert_called_with(self.model.id)
        mock_load.assert_called_with(**self.model._api.get.return_value.__dict__)

    def test_json(self, mock_raise):
        self.assertEqual(self.model.json(), '{"foo": "bar", "id": "abc123"}')

    def test_json_private(self, mock_raise):
        self.model._private = 'foo'
        self.assertEqual(self.model.json(), '{"foo": "bar", "id": "abc123"}')

    def test_json_lazy_loaded(self, mock_raise):
        self.model.test = air_model.LazyLoaded('foo', 'bar')
        self.assertEqual(self.model.json(), '{"foo": "bar", "id": "abc123", "test": "foo"}')

    def test_json_lazy_loaded_list(self, mock_raise):
        self.model.test = air_model.LazyLoadedList([air_model.LazyLoaded('a', 1),
                                                    air_model.LazyLoaded('b', 2)], MagicMock())
        self.assertEqual(self.model.json(), '{"foo": "bar", "id": "abc123", "test": ["a", "b"]}')

    def test_json_datetime(self, mock_raise):
        time = dt.datetime(2030, 12, 12, 22, 5, 3)
        self.model.test = time
        self.assertEqual(self.model.json(),
                         '{"foo": "bar", "id": "abc123", "test": "2030-12-12T22:05:03"}')

class TestLazyLoaded(TestCase):
    def setUp(self):
        self.model = air_model.LazyLoaded('abc123', 'tests')

    def test_init(self):
        self.assertEqual(self.model.id, 'abc123')
        self.assertEqual(self.model.model, 'tests')

    def test_repr(self):
        self.assertEqual(str(self.model), f'<air_sdk.air_model.LazyLoaded Test {self.model.id}>')

    def test_repr_topology(self):
        model = air_model.LazyLoaded('abc123', 'topologies')
        self.assertEqual(str(model), f'<air_sdk.air_model.LazyLoaded Topology {self.model.id}>')

class TestLazyLoadedList(TestCase):
    def setUp(self):
        self.api = MagicMock()
        self.item1 = air_model.LazyLoaded('abc', 'tests')
        self.item2 = air_model.LazyLoaded('xyz', 'tests')
        self.model = air_model.LazyLoadedList([self.item1, self.item2], self.api)

    def test_init(self):
        self.assertEqual(self.model._api, self.api)
        self.assertListEqual(self.model, [self.item1, self.item2])

    def test_getitem(self):
        self.assertEqual(self.model[0], self.api.client.tests.get.return_value)
        self.api.client.tests.get.assert_called_with('abc')

    def test_iter(self):
        mock_item = MagicMock()
        mock_item.test = 'foo'
        self.api.client.tests.get.return_value = mock_item
        for item in self.model:
            self.assertEqual(item.test, 'foo')

class TestHelpers(TestCase):
    def test_get_item_id(self):
        self.assertEqual(air_model._get_item_id('foo'), 'foo')

    def test_get_item_id_dict(self):
        self.assertEqual(air_model._get_item_id({'id': 'abc123'}), 'abc123')

    def test_get_item_id_url(self):
        self.assertEqual(air_model._get_item_id('http://testserver/api/v1/test/abc123'), 'abc123')



# SPDX-FileCopyrightText: Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

""" Setup script """

import setuptools

with open('README.md', 'r') as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name='air-sdk',
    version='2.4.0',
    author='Mike Robertson',
    author_email='mrobertson@nvidia.com',
    description='Python SDK for interacting with NVIDIA Air',
    license='MIT',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/NVIDIA/air_sdk',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    install_requires=['python-dateutil', 'requests']
)
